VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "ParserObj"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Compare Database

'   ParserObj v 1.1.0   - a text parser tool
'   Author:  Richard W Hunt Jr
'   (C) Copyright 2020 - all rights reserved
'   READ ME**********************************************************************
'   License to use this software is granted for no monetary cost.  However anyone
'   who wishes to incorporate this code in a project of their own must include an
'   acknowledgment of the origin of the code.
'   The author offers no warranty, expressed or implied, regarding correctness of
'   the code.  Persons incorporating any or all of the elements of this module are
'   implicitly accepting all risks and responsibilities relating to the use of the
'   module in their own work.
'   Persons wishing to modify the code have permission to do so, but must include
'   full disclosure of such modification.
'   Maintenance of the routine is not guaranteed but can be requested from user
'   The_Doc_Man, a member of the Access World Forums found at web site
'   https://www.access-programmers.co.uk/forums/
'   However, the Access World Forums site bears no responsibility for the code or
'   its correctness.
'*********************************************************************************
'   History
'---------------------------------------------------------------------------------
'   5/2020 - v1.1.0  - Fix ParseFromCol to adjust to start of nearest token and
'           assure alignment of TkPointer for both ParseFromCol and ParseFromOrd
'   4/2020 - v1.0.8  - Clean up error handling.  Implement ParseToPrntChr
'   4/2020 - v1.0.7  - Add TkChrName to translate non-printing characters
'   3/2020 - v1.0.6  - Change class tclEOLChr from multi- to single-character
'   2/2020 - v1.0.5  - Add ParseFromOrd and ParseFromCol
'   12/2019 - v1.0.4 - Add TkPointer for parser support of error-pointer
'   10/2019 - v1.0.3 - Add TkName to provide name of token class
'   9/2019 - v1.0.2  - Add Options for DollarSign and UnderScore.  Fix option for
'           quoted strings.
'   9/2019 - v1.0.1  - Internal routine ChrCode split to separate sub so can use
'           in other than in-line in the Parser step routine
'   8/2019 - v1.0.0  - Original version adapted from an older project
'---------------------------------------------------------------------------------

Private Const cvVersn = "1.1.0"     'the version number of the parser class object

'================================ token type-codes

Public Enum TknCode        'the class-codes of recognized token types

    tclBadChr = -2          'token is a single character we do not use (not US ASCII)
                            'parsing can continue past this character
    tclErr = -1             'token represents an erroneous state (and is likely empty)
                            'parsing should not continue past this condition
    tclNoCls = 0            'token is in "initialized" state; no parsing has occurred yet
                            'parsing after seeing this is perfectly OK
    tclAlpha = 1            'token is mixed upper-case/lower-case alphabetic
                            'can be multiple characters.  SEE ALSO OptAN option
    tclDigits = 2           'token is numeric (and breaks at next alpha)
                            'can be multiple characters
    
    tclDotCma = 3           'token is single dot or comma
    tclClnSemi = 4          'token is single colon or semicolon
    
    tclBrkts = 5            'token is any single bracketing character () {} []
    tclMath = 6             'token is any single arithmetic character + - * /
    tclReln = 7             'token is any relational character < = > and also combos <> <= >=
                            'can be 1 or 2 characters
                            
                            'quoting characters are apostrophe/single quote, double-quote, accent
    tclIsoQuot = 8          'token is any single (isolated) quoting character
                            'isolated = only one such character on the input line
    tclQuotStr = 9          'token is any quoted string using same starting and ending quoting char.
                            'is at least 2 characters plus whatever was enclosed by the quotes
    
    tclGenPunct = 10        'token is any single punctuation - ampersand underscore ! @ # $ % ^ | \ ? ~
    
    tclSpcTab = 11          'token is any sequence of spaces and tabs combined to make "white space"
                            'can be multiple characters
    
    tclEOLCHR = 12          'token is any control character sequence considered as EOL: CR LF FF VT
                            'parsing could potentially continue
    tclGenCtl = 13          'token is any single non-space, non-EOL control character
                            'includes NUL, DEL, SOH, STX, and quite a few more
                            
    tclEOL = 14             'token represents a positional EOL (to right of last character on line)
                            '*** parsing must terminate upon reachng tclEOL as a returned token class
End Enum

'================================ the errors

Public Enum tknError
    PRSERR_NO_ERROR = 0             'no error on record at the moment
    PRSERR_INVAL_USE_OF_NULL = vbObjectError + 1000 'invalid use of null
    PRSERR_PARSE_PAST_EOL = vbObjectError + 1001    'Parse method called when had already reached EOL
    PRSERR_BAD_PARAMETER = vbObjectError + 1002     'Some method called with bad parameter
    PRSERR_PARSE_NO_INPUT = vbObjectError + 1003    'Parse method called before new string loaded
    PRSERR_POSITION_LOST = vbObjectError + 1004     'Internal position pointer lost
    PRSERR_BAD_CALL_CONTEXT = vbObjectError + 1005  'Call tried to do something in the wrong context
End Enum

'================================ the properties

Dim csVersion As String             'the version number of the parser.
Dim csClause As String              'the input line to be parsed (unedited)
Dim csToken As String               'the characters of the most recent token, can be empty
Dim csPosPnt As String              'a blank line (plus a caret) that shows where a parse stopped.

Dim clCLength As Long               'the length of the (input) clause
Dim clTLength As Long               'the length of the most recent token
Dim clTStart As Long                'the starting position of the token
Dim clTOrdnl As Long                'the ordinal position of the token (1st token, 2nd token, etc.)
Dim clNDict As Long                 'number of objects in dictionary

Dim ccTCCode As TknCode             'the class-code of the token
Dim ccTPrCod As TknCode             'the previous token class-code (property internal only, not exposed)

Dim ceErrCode As tknError           'the error code

Dim cbAlphaNum As Boolean           'if true, allow numerics in string that started with alphabetics;
                                    'if false, numerics are always separate class from alphabetics
Dim cbPureAlpha As Boolean          'if true, the current token contains only alphabetics (NOT A PROPERTY)
                                    'if false, the current token contains non-alphabetics (NOT A PROPERTY)
Dim cbQuotedStrOK As Boolean        'if true, quoted strings are allowed.
                                    'if false, all quotes are viewed in isolation.
Dim cbAllowUNDSCR As Boolean        'if true, underscore can be part of an alphanumeric
                                    'if false OR if cbAlphaNum is false, underscore is disallowed in alphas
Dim cbAllowDLRSGN As Boolean        'if true, dollar sign can be part of an alphanumeric
                                    'if false OR if cbAlphaNum is false, dollar sign is disallowed in alphas
Dim cbEOL As Boolean                'if true, we have reached a positional EOL
Dim cbERR As Boolean                'if true, parser has raised an error that should be honored

Dim cdDict As Scripting.Dictionary  'internal dictionary object

Option Explicit

Private Sub Class_Initialize()
'
'   establish initial context for all data elements
'
    csVersion = cvVersn             'immediately assert version number
    cbAlphaNum = False              'by default, alphabetics stay pure
    cbQuotedStrOK = False           'by default, quotes are always isolated
    cbAllowUNDSCR = False
    cbAllowDLRSGN = False
    cbPureAlpha = False             'internal non-property flag
    Set cdDict = CreateObject("Scripting.Dictionary")
    clNDict = 0                     'dictionary is empty
    Class_Reset                     'reset everything else
    
End Sub

Private Sub Class_Terminate()
'
'   right now, a formality that cleans up after itself.  could be more later
'
    Call Class_Reset                'reset everything to minimize space
    cdDict.RemoveAll                'erase the dictionary
    Set cdDict = Nothing            'dereference the dictionary
    
End Sub

Public Sub ResetPrs()
'
'   this allows a manual reset (but does not affect AlphaNum flag)
'   technically not required if you load up a new line because doing
'   so ALSO calls the reset routine.  But hey, it is cheap to implement.
'
    Class_Reset                     'erase remnants of prior parsing
    
End Sub

Private Sub Class_Reset()
'
'   reset everything to either blank or a code for "nothing done"
'   IF you query the class object now, since it has just been reset, you will see
'   that it returns an error because has not parsed anything yet.
'
    csClause = ""                   'empty out the string
    csPosPnt = ""                   'no string? no position pointer either.
    clCLength = 0                   'see... zero length
    csToken = ""                    'if clause is a ZLS, so is token
    clTLength = 0                   'therefore, also zero length
    ccTCCode = tclNoCls             'no parse = no class
    ccTPrCod = tclNoCls
    clTStart = 0                    'technically, to left of first character
    clTOrdnl = 0                    'nothing parsed yet
    cbERR = True                    'asking status right now is bad
    cbEOL = True                    'we cannot parse past this point
    ceErrCode = PRSERR_PARSE_NO_INPUT   'if you call now, this is what comes
    cdDict.RemoveAll                'erase but do not destroy the dictionary
    clNDict = 0                     'dictionary is empty
    
End Sub

Public Property Get Version() As String
'
'   return the version number of this parser class object
'
    Version = csVersion             'the current version of this parser class object
    
End Property

Public Property Get OptQS() As Boolean
'
'   return the state of the "iso quote"/"quoted string" flag
'   if TRUE, allow quoted strings
'   if FALSE, all quote characters are isolated.
'
    OptQS = cbQuotedStrOK
    
End Property

Public Property Get OptAN() As Boolean
'
'   return state of "pure"/"impure" alpha flag.
'   If TRUE, parser allows alphanumerics.
'   If FALSE, parse keeps alpha totally separare from numerics
'
    OptAN = cbAlphaNum           'return state of flag
    
End Property

Public Property Get OptDS() As Boolean
'
'   return state of "allow dollarsign in alphanum" flag
'   if TRUE, parser will allow alphanumerics to also contain dollar sign
'   if FALSE, parer will disallow dollar sign in alphanumerics
'   if AlphaNum is false, state of dollar sign flag is immaterial but
'   nevertheless is reported correctly.
'
    OptDS = cbAllowDLRSGN
    
End Property

Public Property Get OptUS() As Boolean
'
'   return state of "allow underscore in alphanum" flag
'   if TRUE, parser will allow alphanumerics to also contain underscores
'   if FALSE, parser wil disallow underscore in alphanumerics
'   if AlphaNum is false, state of underscore flag is immaterial
'   always reported correctly even if AlphaNum is not set.

    OptUS = cbAllowUNDSCR

End Property

Public Property Let OptAN(TheFlag As Boolean)
'
'   must define the alphanumeric flag before first use or after a direct reset
'   flag argument must not be null
'
    On Error GoTo ErrHand
    If (clTOrdnl <> 0) Or (clCLength <> 0) Then
        cbERR = True
        ceErrCode = PRSERR_BAD_CALL_CONTEXT
        cbAlphaNum = False
'        Err.Raise ceErrCode, TypeName(Me), XLateErrCode(ceErrCode)
    ElseIf IsNull(TheFlag) Then
        cbERR = True
        ceErrCode = PRSERR_INVAL_USE_OF_NULL
        cbAlphaNum = False
'        Err.Raise ceErrCode, TypeName(Me), XLateErrCode(ceErrCode)
    Else
        cbAlphaNum = TheFlag        'set internal flag based on call param
    End If
    
    Exit Property

ErrHand:

    Err.Raise Err.Number, TypeName(Me), XLateErrCode(Err.Number)
    
    Resume Next
    
End Property

Public Property Let OptDS(TheFlag As Boolean)
'
'   assert preference for allowing dollarsigns in alphanumerics
'   flag argument must not be null.  Note that even if this flag is
'   set TRUE, it has no effect if AlphaNum is FALSE.

    On Error GoTo ErrHand
    If (clTOrdnl <> 0) Or (clCLength <> 0) Then     'allowed to do this now?
        cbERR = True                                'no, not allowed
        ceErrCode = PRSERR_BAD_CALL_CONTEXT
        cbAllowDLRSGN = False
    ElseIf IsNull(TheFlag) Then                     'garbage input?
        cbERR = True
        ceErrCode = PRSERR_INVAL_USE_OF_NULL        'yes, complain
        cbAllowDLRSGN = False
    Else
        cbAllowDLRSGN = True                        'just assert the flag
    End If
    
    Exit Property
    
ErrHand:

    Err.Raise Err.Number, TypeName(Me), XLateErrCode(Err.Number)
    
    Resume Next

End Property

Public Property Let OptUS(TheFlag As Boolean)
'
'   assert preference for allowing underscores in alphanumerics.
'   flag argument must not be null.  Even if the flag is TRUE,
'   it has no effect of AlphaNum is false.
'
    On Error GoTo ErrHand
    If (clTOrdnl <> 0) Or (clCLength <> 0) Then
        cbERR = True
        ceErrCode = PRSERR_BAD_CALL_CONTEXT
        cbAllowUNDSCR = False
    ElseIf IsNull(TheFlag) Then
        cbERR = True
        ceErrCode = PRSERR_INVAL_USE_OF_NULL
        cbAllowUNDSCR = False
    Else
        cbAllowUNDSCR = TheFlag
    End If
    
    Exit Property

ErrHand:

    Err.Raise Err.Number, TypeName(Me), XLateErrCode(Err.Number)
    
    Resume Next

End Property

Public Property Let OptQS(TheFlag As Boolean)
'
'   assert preference for allowing quoted strings as an entity.
'   context and content restrictions apply
'
    On Error GoTo ErrHand
    If (clTOrdnl <> 0) Or (clCLength <> 0) Then
        cbERR = True
        ceErrCode = PRSERR_BAD_CALL_CONTEXT
        cbQuotedStrOK = False
'        Err.Raise ceErrCode, TypeName(Me), XLateErrCode(ceErrCode)
    ElseIf IsNull(TheFlag) Then
        cbERR = True
        ceErrCode = PRSERR_INVAL_USE_OF_NULL
        cbQuotedStrOK = False
'        Err.Raise ceErrCode, TypeName(Me), XLateErrCode(ceErrCode)
    Else
        cbQuotedStrOK = TheFlag
    End If
    
    Exit Property

ErrHand:

    Err.Raise Err.Number, TypeName(Me), XLateErrCode(Err.Number)
    
    Resume Next
    
End Property

Public Property Get TkErrCode() As tknError
'
'   return the error code number
'
    TkErrCode = ceErrCode
    
End Property

Public Property Get TkErr() As Boolean
'
'   return the current error code.  no processing, just copy it.
'
    TkErr = cbERR                   'return current error flag as T/F
    
End Property

Private Function XLateErrCode(ecode As tknError) As String
'
'   provide a text meaning for the errors
'
    Select Case ecode
        Case PRSERR_NO_ERROR
            XLateErrCode = "No error"
        Case PRSERR_INVAL_USE_OF_NULL
            XLateErrCode = "Invalid use of null"
        Case PRSERR_PARSE_PAST_EOL
            XLateErrCode = "Parse past end of input line"
        Case PRSERR_BAD_PARAMETER
            XLateErrCode = "Bad parameter"
        Case PRSERR_PARSE_NO_INPUT
            XLateErrCode = "Parser has no input"
        Case PRSERR_POSITION_LOST
            XLateErrCode = "Parser position lost"
        Case PRSERR_BAD_CALL_CONTEXT
            XLateErrCode = "Call made in bad context"
        Case Else
            XLateErrCode = "Internal error - unknown error code " & CStr(ceErrCode)
    End Select

End Function

Public Property Get TkErrStr() As String
'
'   return the error code as a text string
'
    TkErrStr = XLateErrCode(ceErrCode)
    
End Property

Public Property Get TkNum() As Long
'
'   return the token ordinal position.  0 means no parsing has occurred.
'
    TkNum = clTOrdnl             'return ordinal position of this token (0=no parse yet)
    
End Property

Public Property Get Token() As String
'
'   return the actual token string (unedited)
'
    Token = csToken                 'this is the token returned by a parse step
    
End Property

Public Property Get TkSize() As Long
'
'   return the size of the current token.  CAN BE ZERO in some cases.
'
    TkSize = clTLength           'return size of current token (EOL, ERR, and NOCLS can return 0)
    
End Property

Public Property Get TkStart() As Long
'
'   return the starting position of the current token
'
    TkStart = clTStart           'character position of start of token from input line
    
End Property

Public Property Get TkCode() As TknCode
'
'   return the current token code.  INIT code sets this to a valid state so no error checking needed.
'
    TkCode = ccTCCode            'report the token class-code
    
End Property

Public Property Get AtEOL() As Boolean
'
'   return state of EOL flag
'
    AtEOL = cbEOL                   'positional EOL

End Property

Public Property Let ParseLine(TheLine As String)
'
'   prepare for parsing by loading up a string.  string can contain up to several thousand characters.
'   once all is reset, we are ready to actually load things up
'   measure line size, reset error code.  Token class-code is reset by Class_Reset call.
'
    On Error GoTo ErrHand
    If IsNull(TheLine) Then
        cbERR = True
        ceErrCode = PRSERR_INVAL_USE_OF_NULL
'        Err.Raise ceErrCode, TypeName(Me), XLateErrCode(ceErrCode)
    Else
        Call Class_Reset                'erase all remnants.  resets almost everything.
        csClause = TheLine              'load up the string
        clCLength = Len(csClause)       'set the internal line length
        cbEOL = False                   'new line so not at EOL yet
        cbERR = False                   'and give benefit of the doubt at first
        ccTCCode = tclNoCls             'no parse has yet occurred so mark consistent with that fact
        ceErrCode = PRSERR_NO_ERROR     'no errors at the moment
        csPosPnt = ""                   'no error pointer
    End If

    Exit Property

ErrHand:

    Err.Raise Err.Number, TypeName(Me), XLateErrCode(Err.Number)
    
    Resume Next

End Property

Public Property Get ParseLine() As String
'
'   return the string with which we are currently working
'
    ParseLine = csClause

End Property

Public Property Get TkPointer() As String
'
'   return the accumulating error pointer line
'
    TkPointer = csPosPnt         'just copy it
    
End Property

Public Property Get LnSize() As Long
'
'   return length of the line (or at least, how we see it...)
'
    LnSize = clCLength            'length of the input line or clause

End Property

Public Property Get TkName() As String
'
'   function returns the name of the current token class based on ccTCCode
'
    Select Case ccTCCode            'based on the class, return a name
        Case tclNoCls
            If clTOrdnl = 0 Then
                TkName = "No Parse"
            Else
                TkName = "Undetermined"
            End If
        Case tclAlpha
            If cbAlphaNum Or (Not cbPureAlpha) Then
                TkName = "Alphanumeric"
            Else
                TkName = "Alphabetic"
            End If
        Case tclDigits
            TkName = "Digits"
        Case tclDotCma
            TkName = "Dot/Comma"
        Case tclClnSemi
            TkName = "Colon/Semicolon"
        Case tclBrkts
            TkName = "Bracketing Symbols"
        Case tclMath
            TkName = "Math Operator"
        Case tclReln
            TkName = "Relational Operator"
        Case tclIsoQuot
            TkName = "Quoting character"
        Case tclQuotStr
            TkName = "Quoted string"
        Case tclGenPunct
            TkName = "General Punctuation"
        Case tclSpcTab
            TkName = "White Space (Spaces/Tabs)"
        Case tclEOLCHR
            TkName = "End of Line (characters)"
        Case tclGenCtl
            TkName = "General control character"
        Case tclBadChr
            TkName = "Non-ASCII character"
        Case tclEOL
            TkName = "End of Line (position)"
        Case Else
            TkName = "Error"
    End Select

End Property

Public Sub RestartPrs()
'
'   call this to reset the current line as though we were back before the first ParseNext call
'
    On Error GoTo ErrHand
    csToken = ""                    'restarted token is always ZLS
    clTLength = 0                   'therefore, reflect zero length
    ccTPrCod = tclNoCls             'no previous class because starting over.
    clTStart = 0                    'technically, to left of first character
    clTOrdnl = 0                    'nothing parsed yet
    csPosPnt = ""                   'reset error stream pointer
    cdDict.RemoveAll                'erase the dictionary
    clNDict = 0                     'nothing in dictionary
    
    If clCLength > 0 Then           'do we have a line defined?
        cbEOL = False               'not at end of line
        ceErrCode = PRSERR_NO_ERROR 'and this is a no-error situation
        cbERR = False               'normal situation
        ccTCCode = tclNoCls         'no parse = no class
    Else
        cbEOL = True                'we cannot parse past this point
        ceErrCode = PRSERR_PARSE_NO_INPUT   'if you call now, this is what comes
        cbERR = True                'bad situation
        ccTCCode = tclEOL           'nothing to parse
'        Err.Raise ceErrCode, TypeName(Me), XLateErrCode(ceErrCode)
    End If

    Exit Sub

ErrHand:

    Err.Raise Err.Number, TypeName(Me), XLateErrCode(Err.Number)
    
    Resume Next

End Sub

Private Function ChrCode(sChar As String) As TknCode
'
'   given the leftmost input character, determine its class-code and return that to the
'   caller (which is only the parser at this time)
'
'   note that the order of this case-ladder is NOT sorted by ASCII number but by most
'   likely frequency of occurrence of the given characters
'

Dim lAsc As Long                    'a place for the ASCII character numeric translation

    lAsc = Asc(Left(sChar, 1))      'extract leftmost (1st) character of input
    
    Select Case lAsc
                                        'white spaces most likely
        Case 32: ChrCode = tclSpcTab    '=> space
        Case 9: ChrCode = tclSpcTab     '=> HTAB
        
                                        'upper/lower case letters most likely
        Case 65 To 90: ChrCode = tclAlpha   '=> ABCDEFGHIJKLMNOPQRSTUVWXYZ
        Case 97 To 122: ChrCode = tclAlpha  '=> abcdefghijklmnopqrstuvwxyz
        
                                        'digits next most likely
        Case 48 To 57: ChrCode = tclDigits  '=> digits 0123456789 (not in that order)
        
                                        'certain punctuation more common in SQL usage
        Case 33: ChrCode = tclGenCtl    '=> bang (!)
        Case 46: ChrCode = tclDotCma    '=> period or {.} or dot
        Case 44: ChrCode = tclDotCma    '=> comma or (,)
        Case 58 To 59: ChrCode = tclClnSemi '=> colon (:) or semicolon (;)
        
                                        'quoting characters next
        Case 39: ChrCode = tclIsoQuot   '=> single-quote or apostrophe
        Case 34: ChrCode = tclIsoQuot   '=> double-quote
        Case 96: ChrCode = tclIsoQuot   '=> accent or (`)
        
                                        'parenthese and brackets fairly common
        Case 40 To 41: ChrCode = tclBrkts   '=> ( ) or parentheses - bracketing characters
        Case 91: ChrCode = tclBrkts     '=> open square bracket ([)
        Case 93: ChrCode = tclBrkts     '=> close square bracket (])
        
                                        'math operators and relationals are seen sometimes
        Case 42 To 43: ChrCode = tclMath '=> math operators * and +
        Case 45: ChrCode = tclMath      '=> math operator - or minus or dash
        Case 47: ChrCode = tclMath      '=> slash or (/) or virgule
        Case 92: ChrCode = tclGenPunct  '=> backslant or (\) or reverse virgule
        Case 60 To 62: ChrCode = tclReln '=> relational operators (<) (=) (>)
        
                                         'the rest of these are fairly unlikely
        Case 35 To 38: ChrCode = tclGenPunct   '=> # $ % ampersand
        Case 63 To 64: ChrCode = tclGenPunct '=> question mark (?) and at-sign (@)
        Case 94 To 95: ChrCode = tclGenPunct    '=> caret or (^) or hat, and underscore (_)
        Case 123: ChrCode = tclBrkts     '=> left brace or ({)
        Case 124: ChrCode = tclGenPunct  '=> vertical bar (|)
        Case 125: ChrCode = tclBrkts     '=> right brace or (})
        Case 126: ChrCode = tclGenPunct  '=> tilde or (~)
        Case 10 To 13: ChrCode = tclEOLCHR  '=> LF VT FF CR
        Case 127: ChrCode = tclGenCtl    '=> DEL character
        Case 0 To 8: ChrCode = tclGenCtl '=> NUL SOH STX ETX EOT ENQ ACK BEL BS
        Case 14 To 31: ChrCode = tclGenCtl  '=> SO SI DLE DC1 DC2 DC3 DC4 NAK SYN ETB CAN EM SUB ESC FS GS RS US
        
       Case Else: ChrCode = tclBadChr      '=> not an ASCII character
       
    End Select

End Function

Public Sub ParseNext()
'
'   look at the current information in the class data area.  parse the next item if possible.
'

Dim lTPos As Long                   'position of first character of current token
Dim lCPos As Long                   'position of current character of current token
Dim lAPos As Long                   'auxiliary testing position for special-case testing
Dim lTSiz As Long                   'token size of currently accumulating token

Dim cCCod As TknCode                'code for current character
Dim cTCod As TknCode                'code for token being accumulated

Dim bKeepOn As Boolean              'flag:  Keep parsing
Dim bDropLC As Boolean              'flag:  Drop last character (cannot keep it)
Dim bNoMove As Boolean              'flag:  empty token involved

Dim sAChr As String                 'character used for comparison/testing
Dim sCChr As String                 'current character being considered for similarities

'   test for initial conditions that would stop us dead before we even started

    On Error GoTo ErrHand           'make this almost non-stop
    
'   special case #1 = parsing past EOL character triggers specific error

    If cbEOL = True Then            'trying to parse past EOL?
        ceErrCode = PRSERR_PARSE_PAST_EOL   'signal this one
        GoTo PNFail
    End If
    
'   special case #2 = parsing after prior error noted.  Error flag is already true.
'   triggers specific error.
    
    If cbERR = True Then            'trying to parse past an error?
        ceErrCode = PRSERR_BAD_CALL_CONTEXT 'signal this error
        GoTo PNFail
    End If
    
'   special case #3 = pointer is totally outside the confines of the target line.
'   This represents some sort of internal error that should never ever happen...
'   ...but if it ever does, we have to stop it cold.  Triggers specific error

    If (clTStart > clCLength) Or (clTStart < 0) Then
        ceErrCode = PRSERR_POSITION_LOST   'pointers are impossible
        GoTo PNFail
    End If
        
'
'   prepare internal variables for parsing...
'
    sCChr = ""                      'current character is a ZLS for now
    cCCod = tclNoCls                'make no assumptions
    cTCod = tclNoCls                'none at all
    lCPos = clTStart                'starting from here (from master copy of start point)
    lTPos = lCPos + 1               'probable start of next token (if not EOL)
    lTSiz = 0                       'nothing accumulated yet for current token
    bNoMove = False                 'assume normal movement
    bDropLC = False                 'assume keep last character
    bKeepOn = True                  'assume we can continue to parse
    
    Do While bKeepOn = True         'scan until we can scan no more
        lCPos = lCPos + 1           'step to next position
        lTSiz = lTSiz + 1           'and count it
        
        If lCPos > clCLength Then   'does this step go past end of line?
            cCCod = tclEOL          'yes - positional end of line (ChrCode cannot return this)
        Else
            sAChr = sCChr           'remember previous character
            sCChr = Mid(csClause, lCPos, 1) 'extract next character
            cCCod = ChrCode(sCChr)  'evaluate this character
        End If                      'end of - if just crossed over EOL
        
'   We have an assumed or discovered "previous" code and a character code for current character
'   compare previous and current codes.  If they match, we are compatible and can move forward.
'   If they do not match, we have a few cases.
'   As a quirk, allow underscore as 2nd or later char in an alphabetic.
        
        If cCCod <> cTCod Then      'is this an exact match?
            Select Case cTCod       'no, see what we are accumulating
            
                Case tclNoCls       'prior class was dummy code (i.e. this is 1st character)
                    cTCod = cCCod   'transfer character code to become new token code - continue
                                    'this is where tclEOL is actually declared when it happens.
                    cbPureAlpha = (cCCod = tclAlpha)    'assume this will be pure alphabetic
                    
                Case tclAlpha       'token started as ALPHABETIC but current character is not Alpha
                    If cbAlphaNum = False Then  'if disallowing alphanumeric, leave now for sure
                        bKeepOn = False     'disallowing mixed alphanum
                        bDropLC = True      'cannot keep this character
                    Else            'next IF lets through digits and dollar signs and underscores
                        If cCCod <> tclDigits Then  'if not digits then check for underscore
                            If sCChr = "$" And cbAllowDLRSGN = True Then
                                cbPureAlpha = False        'allow it, but mark as impure
                            ElseIf sCChr = "_" And cbAllowUNDSCR = True Then
                                cbPureAlpha = False        'allow but mark as impure
                            Else
                                bKeepOn = False     'got to stop here
                                bDropLC = True      'and cannot keep this
                            End If  'end of - if not an underscore
                        Else        'ELSE of IF DIGITS
                            cbPureAlpha = False
                        End If      'end of - if not digits
                    End If          'end of - if Alphanum not allowed
                Case Else           'all else will be single-character cases anyway
                    bKeepOn = False 'this breaks the sequence
                    bDropLC = True  'so cannot keep it
            End Select
        End If

'   previous complex IF /SELECT took care of alphanumeric test and 1st-character-in-token case
'   now we have to look at what the character class actually was because of special effects
'   the Alpha, Digits, SpcTab, and EOLChr cases can accumulate stuff.  But test for special other
'   cases that can accumulate - like relationals (1 or 2 characters), quoted strings (can be long)
'   Also trap EOL (Positional) case here to assure we have a stopper for it.
'   The ones that can accumulate are called out but do nothing because the CASE ELSE will stop any
'   tokens that were not otherwise handled already

        Select Case cTCod               'if there is a special effect here, it takes effect here
            Case tclAlpha               'alpahbetic or alphameric - allow sequences
            
            Case tclDigits              'digit sequence - allow sequences
            
            Case tclSpcTab              'white space - allow sequences
            
            Case tclEOL                 'positional EOL - have special action
                bNoMove = True          'do not move for this one
                bKeepOn = False         'it is also a scan-stopper
                
            Case tclReln                'allow single = < or > and either <> or >= or <=
                Select Case lTSiz       'want special case tests
                    Case 1              'allow 1-character relational unconditionally
                    Case 2
                        If sAChr = "=" Then     'check for "=x" case - never allowed
                            bKeepOn = False     'cannot keep going
                            bDropLC = True      'drop this one, cannot keep it
                        ElseIf sAChr = ">" Then 'check for ">=" case
                            If sCChr <> "=" Then    'if not "=" then disallow
                                bKeepOn = False
                                bDropLC = True
                            End If
                        Else            'this HAS to be sAChr = "<"
                            If sCChr = "<" Then     'allow either <> or <=, not <<
                                bKeepOn = False     'stop it here
                                bDropLC = True      'cannot keep it
                            End If
                        End If
                    Case Else           '3 or more relational characters?  yecch!
                        bKeepOn = False 'never allow triple-character relationals
                        bDropLC = True  'cannot keep last one
                End Select
                
            Case tclIsoQuot             'see if this could possibly be start of quoted string?
                lAPos = InStr(lCPos + 1, csClause, sCChr)   'is there another quote like this to the left?
                If (lAPos = 0) Or (cbQuotedStrOK = False) Then  'could it be a quoted string?
                    bKeepOn = False     'either not balanced or quoted string disallowed
                Else
                    lTSiz = (lAPos - lTPos) + 1 'size of quoted string (including quotes)
                    lCPos = lAPos       'reset position
                    bKeepOn = False     'stop with whole string
                    cTCod = tclQuotStr  'this is a quoted string, not an isolated quote-mark
                End If                  'end of - if could be a quoted string case
                
            Case Else
                bKeepOn = False         'anything else here will be single-character return
                                        'i.e. mostly control characters and punctuation
        End Select
        
    Loop                                'loop ends when bKeepOn becomes false
    
    GoTo PNDone                         'step over common-fail code
    
'   Parsing failed for some reason.  Code is in ceErrCode, we do not care where we are.
    
PNFail:
        
        cTCod = tclErr              'error in parsing returns error code...
        cbERR = True                '...and sets the error property
        
        bKeepOn = False             'we are stuck here, must stop
        bNoMove = True              'must not move from where we are
        bDropLC = False             'not dropping anything - stay right here as-is
        
'        Err.Raise ceErrCode, TypeName(Me), XLateErrCode(ceErrCode)
'
'   parsing step completed, 3 possible cases:
'       1   No Move (a zero-length token, could be error or EOL)
'       2   Drop Last (token not empty but last char. tested not compatible)
'       3   Keep All (token not empty, keeping what we last saw)
'
PNDone:

    If bNoMove = True Then          'this generally stops us
        clTLength = 0               'output token is empty
        csToken = ""                'a zero-length string
    Else
        If bDropLC = True Then      'dropping the last character?
            lTSiz = lTSiz - 1       'yes, cut it short
        End If                      'END IF dropping last character
        
        clTOrdnl = clTOrdnl + 1         'remember token ordinal position on this line.
        cdDict.Add clTOrdnl, clTStart   'remember starting position of this token
        clNDict = clTOrdnl          'remember highest defined dictionary slot
    
        clTLength = lTSiz           'transfer size
        clTStart = clTStart + lTSiz 'boost previous start by size of this one
        csToken = Mid(csClause, lTPos, lTSiz)   'extract copy from line
        
        If cTCod = tclSpcTab Then   'returning spaces or tabs?
            csPosPnt = csPosPnt & csToken   'yes, add into error pointer line
        Else
            csPosPnt = csPosPnt & Space$(lTSiz) 'no, provide spaces for error line
        End If
        
    End If                          'END - if not moving (error/stopper)
    
    cbEOL = (ccTCCode = tclEOL)     'if this was EOL, set flag at this time
    ccTCCode = cTCod                'assert this as returned class code
    
    Exit Sub                        'we are done, gedoudahere

ErrHand:

    Err.Raise Err.Number, TypeName(Me), XLateErrCode(Err.Number)
    
    Resume Next
    
End Sub

Public Sub ParseFromOrd(lOrd As Long)

Dim lTPos As Long
Dim lENum As Long

    On Error GoTo ErrHand
    
    If (lOrd < 1) Or (lOrd > clNDict) Then  'see if impossible ordinal
        cbERR = True                'not within boundaries
        ceErrCode = PRSERR_BAD_PARAMETER    'disallow the call
'        Err.Raise ceErrCode, TypeName(Me), XLateErrCode(ceErrCode)
    Else
        lTPos = cdDict.Item(lOrd)   'copy the position
        clNDict = lOrd              'reset dictionary limits
        clTStart = lTPos            'reset parser starting spot
        csPosPnt = Left(csPosPnt, lTPos)    'reset position pointer
        ParseNext                   'parse and update everything
    End If
    
PFO_Exit:

    Exit Sub

ErrHand:

    lENum = Err.Number              'see what trapped us
    Select Case lENum               'dispatch on the number
        Case PRSERR_INVAL_USE_OF_NULL To PRSERR_BAD_CALL_CONTEXT
            ceErrCode = lENum       'one of ours?  Pass it through
        Case Else
            MsgBox "Error " & CStr(lENum) & ":" & Err.Description, vbOKOnly, "Error in Parser Object:ParseFromOrd"
    End Select
    
    cbERR = True
    Resume PFO_Exit

End Sub

Public Sub ParseFromCol(lPos As Long)

Dim lTOrd As Long                               'ordinal index for lookup
Dim lFPos As Long                               'position found from dictionary
Dim lFOrd As Long                               'ordinal found from dictionary
Dim lENum As Long                               'error number for traps

    On Error GoTo ErrHand

'   protect against impossible column number

    If (lPos < 1) Or (lPos > clCLength) Then    'if impossible column, ...
        cbERR = True                            'flag it as an error
        ceErrCode = PRSERR_BAD_PARAMETER        'signal as such
        GoTo PFC_Exit                           'skip on out of here
    End If
    
    If lPos > clTStart Then                     'trying to skip something?
        cbERR = True                            'cannot skip ahead of current parse
        ceErrCode = PRSERR_BAD_PARAMETER        'so counts as a bad column number
        GoTo PFC_Exit                           'short-cut out of here
    End If
    
'   find ordinal containing this column

    lFOrd = 0                                   'make no assumptions
    For lTOrd = 1 To clNDict                    'look for closest token
        lFPos = cdDict.Item(lTOrd)              'find starting column for this ordinal
        If lFPos > lPos Then                    'does this token start beyond selected column?
            Exit For                            'stop looking now, this starts after the col.
        Else
            lFOrd = lTOrd                       'remember the ordinal, still in range
        End If
    Next lTOrd                                  'end of loop regardless of why
    
'   at this point, lFOrd is the ordinal of the token containing the selected column

    If lFOrd = 0 Then                           'started with nothing?
        lFPos = 1                               'yes, reset to start of line
    Else
        lFPos = cdDict.Item(lFOrd)              'get column for this token number
    End If
    
    clNDict = lFOrd                             'reset the dictionary population counter
    clTStart = lFPos                            'and reset starting point for next parse
    csPosPnt = Left(csPosPnt, lFPos)            'trim the position pointer
    ParseNext                                   'parse (all else is done in that code)
    
PFC_Exit:

    Exit Sub

ErrHand:

    lENum = Err.Number                          'see what trapped us
    Select Case lENum                           'dispatch on the number
        Case PRSERR_INVAL_USE_OF_NULL To PRSERR_BAD_CALL_CONTEXT
            ceErrCode = lENum                   'one of ours?  Pass it through
        Case Else
            MsgBox "Error " & CStr(lENum) & ":" & Err.Description, vbOKOnly, "Error in Parser Object:ParseFromCol"
    End Select
    
    cbERR = True
    Resume PFC_Exit
    
End Sub

Public Sub ParseToPrntChr()
'
'   parse until we find a printable character or a parser-stop condition.
'   if possible, skip spaces, tabs, control characters.
'   stop for printable character, parse error, bad character, or EOL
'
PTPC_Go:

    Select Case ccTCCode                    'check for initial stoppers
        Case tclEOL, tclBadChr, tclErr      'any of these and we are done
            GoTo PTPC_End
        Case Else                           'anything else allows at least one more parse
            ParseNext                       'take one step
            Select Case ccTCCode            'see what we got
                Case tclSpcTab, tclEOLCHR, tclGenCtl    'any of these can be skipped
                    GoTo PTPC_Go            'go back, verify and try again
                Case Else                   'everything else falls through
            End Select                      'END inner select
    End Select                              'END outer select
        
PTPC_End:

    Exit Sub                                'return with what we have

End Sub

Public Property Get TkChrName() As String
'
'   call TkChrName to get the name of the first character of the token
'   works even for multi-character tokens but only ever gives name of
'   left-most character in token.  Most useful for non-printing tokens
'
Dim lCode As Long

    If clTLength < 1 Then
        TkChrName = ""
        Exit Property
    End If
    
    lCode = Asc(Left(csToken, 1))
    Select Case lCode
        Case 0
            TkChrName = "<NUL>"
        Case 1
            TkChrName = "<SOH>"
        Case 2
            TkChrName = "<STX>"
        Case 3
            TkChrName = "<ETX>"
        Case 4
            TkChrName = "<EOT>"
        Case 5
            TkChrName = "<ENQ>"
        Case 6
            TkChrName = "<ACK>"
        Case 7
            TkChrName = "<BEL>"
        Case 8
            TkChrName = "<BS>"
        Case 9
            TkChrName = "<HT>"
        Case 10
            TkChrName = "<LF>"
        Case 11
            TkChrName = "<VT>"
        Case 12
            TkChrName = "<FF>"
        Case 13
            TkChrName = "<CR>"
        Case 14
            TkChrName = "<SO>"
        Case 15
            TkChrName = "<SI>"
        Case 16
            TkChrName = "<DLE>"
        Case 17
            TkChrName = "<DC1>"
        Case 18
            TkChrName = "<DC2>"
        Case 19
            TkChrName = "<DC3>"
        Case 20
            TkChrName = "<DC4>"
        Case 21
            TkChrName = "<NAK>"
        Case 22
            TkChrName = "<SYN>"
        Case 23
            TkChrName = "<ETB>"
        Case 24
            TkChrName = "<CAN>"
        Case 25
            TkChrName = "<EM>"
        Case 26
            TkChrName = "<SUB>"
        Case 27
            TkChrName = "<ESC>"
        Case 28
            TkChrName = "<FS>"
        Case 29
            TkChrName = "<GS>"
        Case 30
            TkChrName = "<RS>"
        Case 31
            TkChrName = "<US>"
        Case 32
            TkChrName = "<SP>"
        Case 127
            TkChrName = "<DEL>"
        Case Else
            TkChrName = Left(csToken, 1)        'not a control character, return literal character
    End Select

End Property

